unit TreeScan;

{
Description:

Component to help with traversing folder structures
Will notify the owner when file or folder is found.

Note that files are scanned before sub-folders, and that
both file and folder scans are in alphabetical order

Revision history:

1996 Dec 15  V1.0.0  First version
1997 Jan 12  V1.0.2  Add file mask (e.g. *.exe)
1997 Mar 28  V1.0.4  Add event when leaving folder
                     Add BeginUpdate/EndUpdate around list updates
1997 Apr 26  V1.0.6  Add report of found object size (for files)
                     Add Recursive flag - normally True
                     (suggestions of Sean Mathews)
}

interface

uses
  SysUtils, Classes, Forms;              // seems to be the minimum we need

type
  TTreeScanner = class(TComponent)
  private
    { Private declarations }
    FOnFileFound: TNotifyEvent;          // where to notify for file found
    FOnDirectoryFound: TNotifyEvent;     // where to notify for start of folder
    FOnDirectoryDone: TNotifyEvent;      // where to notify for end of folder
    FFoundObject: string;                // full file or folder name
    FFoundObjectSize: integer;           // size of file found
    FInitialDirectory: string;           // starting point in the tree
    FFileMask: string;                   // files to search for e.g. *.EXE
    FSearchInProgress: boolean;          // true whilst searching
    FRecursive: boolean;
  protected
    { Protected declarations }
    procedure scan_directory (const dir: string);
  public
    { Public declarations }
    Continue: Boolean;                   // caller sets the false to interrupt
    constructor Create (AOwner: TComponent);  override;
    procedure ScanTree;                  // call here to scan tree
  published
    { Published declarations }
    property OnDirectoryFound: TNotifyEvent read FOnDirectoryFound write FOnDirectoryFound;
    property OnDirectoryDone: TNotifyEvent read FOnDirectoryDone write FOnDirectoryDone;
    property OnFileFound: TNotifyEvent read FOnFileFound write FOnFileFound;
    property FileFound: string read FFoundObject;
    property FileSize: integer read FFoundObjectSize;
    property SearchInProgress: boolean read FSearchInProgress;
    property InitialDirectory: string read FInitialDirectory write FInitialDirectory;
    property FileMask: string read FFileMask write FFileMask;
    property Recursive: boolean read FRecursive write FRecursive default True;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Davids', [TTreeScanner]);
end;

constructor TTreeScanner.Create (AOwner: TComponent);
begin
  Inherited Create (AOwner);
  FOnDirectoryFound := nil;       // no callback pointer yet for folder start
  FOnDirectoryDone := nil;        // or for folder complete
  FOnFileFound := nil;            // or for found file
  FSearchInProgress := False;     // and we're not yet searching
  FRecursive := True;             // normally, scan sub-folders as well
end;

procedure TTreeScanner.ScanTree;
begin
  FSearchInProgress := True;      // now searching
  FFoundObject := '';             // nothing found as yet
  FFoundObjectSize := 0;

  // Try to get a valid folder string.  If the user hasn't
  // specified a folder, use the current directory...
  if FInitialDirectory = '' then FInitialDirectory := GetCurrentDir;

  // ensure the folder ends in a '\'
  if FInitialDirectory [Length (FInitialDirectory)] <> '\' then
    FInitialDirectory := FInitialDirectory + '\';

  // now remove any file specification components...
  FInitialDirectory := ExtractFileDir (FInitialDirectory);
  // and again ensure that trailing backslash is present....
  if FInitialDirectory [Length (FInitialDirectory)] <> '\' then
    FInitialDirectory := FInitialDirectory + '\';
  // assume we wish to continue (at least to start with!)
  Continue := True;
  // and call the recursive scanning procedure
  scan_directory (FInitialDirectory);
  FSearchInProgress := False;      // note we have finished searching
end;

procedure TTreeScanner.scan_directory (const dir: string);
var
 f: TSearchRec;             // used for scanning a folder
 status: integer;           // value returned from FindFirst/FindNext
 file_list: TStringList;    // sorted list of files in the folder
 dir_list: TStringList;     // sorted list of sub-folders
 i: integer;                // used to traverse the above lists
begin
  Application.ProcessMessages;    // allow the caller to do some work
  if not Continue then Exit;      // interrupted? - simply return to caller

  // a folder doesn't actually have zero size - it's a file after all
  // but the size isn't reported by FindFirst so say it's zero.
  FFoundObjectSize := 0;             // "a folder has zero size"
  FFoundObject := dir;               // prepare to tell owner that we've
  if Assigned (FOnDirectoryFound)    // found a new folder
    then FOnDirectoryFound (self);

  // Create the string lists which are used to store the names returned from
  // the folder scan.  By making the list sorted, the names returned to the
  // caller will be in alphabetic order.  Note one slight "trick"; the most
  // wanted characteristic of a file is its size.  We therefore store the
  // size in the string list by using its associated "objects" list.  In
  // reality, these are 32-bit pointers to objects.  So rather than storing
  // pointers, we can store and recover 32-bit integers by type-casting.
  // In Win32, a file size is actually a 64-bit field, although Borland only
  // store the low 32-bits.  Once file sizes exceed 2GB this code will break.

  file_list := TStringList.Create;   // prepare list of files
  file_list.Sorted := True;          // and note is is to be in sorted order
  dir_list := TStringList.Create;    // prepare list of folders
  dir_list.Sorted := True;           // also sorted

  dir_list.BeginUpdate;              // probably not worth doing this, but...
  status := FindFirst (dir + '*.*', faDirectory, f);  // try the first find call
  while status = 0 do with f do      // keep looping while more entries
    begin
    if ((faDirectory and Attr) <> 0) and     // is this a folder?
       (Name <> '.') and (Name <> '..')      // ignore backlinks
      then dir_list.Add (dir + Name + '\');  // add it to folder list
    status := FindNext (f);          // see if there's another folder entry
    end;
  FindClose (f);                     // all done, clean up from folder scan
  dir_list.EndUpdate;

  file_list.BeginUpdate;
  status := FindFirst (dir + FFileMask, faAnyFile, f);  // try the first find call
  while status = 0 do with f do      // keep looping while more entries
    begin
    if ((faDirectory and Attr) = 0) and   // is this not a folder?
       ((faVolumeID and Attr) = 0)        // is this not a volume ID?
      then file_list.AddObject (dir + Name, Pointer (Size));    // then add to file list
    status := FindNext (f);          // see if there's another entry
    end;
  FindClose (f);                     // all done, clean up from the scan
  file_list.EndUpdate;

  // scan the sorted list of files and call back the owner for each
  for i := 0 to file_list.Count - 1 do
    begin
    FFoundObject := file_list.Strings [i];
    // type cast the stored pointer into a 32-bit integer
    FFoundObjectSize := Integer (file_list.Objects [i]);
    if Assigned (FOnFileFound) then FOnFileFound (self);
    end;
  file_list.Free;                     // return memory owned by the list

  FFoundObjectSize := 0;              // just to keep things clean.....
  // scan the sorted list of folders and scan each sub-folder
  for i := 0 to dir_list.Count - 1 do
    if FRecursive
    then
      scan_directory (dir_list.Strings [i])
    else
      begin
      FFoundObject := dir_list.Strings [i];
      if Assigned (FOnDirectoryFound) then FOnDirectoryFound (self);
      end;

  dir_list.Free;                     // return memory owned by the list

  if Assigned (FOnDirectoryDone)     // finished this folder
    then FOnDirectoryDone (self);
end;

end.
